import datetime
from pathlib import Path

import pytest
import securesystemslib
import tuf

import taf.exceptions
import taf.yubikey as yk
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.tests import TEST_WITH_REAL_YK
from taf.tests.yubikey_utils import VALID_PIN
from taf.utils import to_tuf_datetime_format


@pytest.mark.skipif(TEST_WITH_REAL_YK, reason="Testing with real Yubikey.")
def test_check_no_key_inserted_for_targets_should_raise_error(repositories, targets_yk):
    taf_happy_path = repositories["test-happy-path"]
    targets_yk.insert()
    targets_yk.remove()
    with pytest.raises(taf.exceptions.YubikeyError):
        taf_happy_path.is_valid_metadata_yubikey("targets")


def test_check_targets_key_id_for_targets_should_return_true(repositories, targets_yk):
    taf_happy_path = repositories["test-happy-path"]
    targets_yk.insert()
    assert taf_happy_path.is_valid_metadata_yubikey("targets", targets_yk.tuf_key)


def test_check_root_key_id_for_targets_should_return_false(repositories, root1_yk):
    taf_happy_path = repositories["test-happy-path"]
    root1_yk.insert()
    assert not taf_happy_path.is_valid_metadata_yubikey("targets", root1_yk.tuf_key)


def test_update_snapshot_valid_key(repositories, snapshot_key):
    taf_happy_path = repositories["test-happy-path"]
    start_date = datetime.datetime.now()
    interval = 1
    expected_expiration_date = to_tuf_datetime_format(start_date, interval)
    targets_metadata_path = Path(taf_happy_path.metadata_path) / "targets.json"
    old_targets_metadata = targets_metadata_path.read_bytes()
    taf_happy_path.update_snapshot_keystores(
        [snapshot_key], start_date=start_date, interval=interval
    )
    new_snapshot_metadata = str(Path(taf_happy_path.metadata_path) / "snapshot.json")
    signable = securesystemslib.util.load_json_file(new_snapshot_metadata)
    tuf.formats.SIGNABLE_SCHEMA.check_match(signable)
    actual_expiration_date = signable["signed"]["expires"]

    # Targets data should remain the same
    assert old_targets_metadata == targets_metadata_path.read_bytes()
    assert actual_expiration_date == expected_expiration_date


def test_update_snapshot_wrong_key(repositories, timestamp_key):
    taf_happy_path = repositories["test-happy-path"]
    with pytest.raises(taf.exceptions.InvalidKeyError):
        taf_happy_path.update_snapshot_keystores([timestamp_key])


def test_update_timestamp_valid_key(repositories, timestamp_key):
    taf_happy_path = repositories["test-happy-path"]
    start_date = datetime.datetime.now()
    interval = 1
    expected_expiration_date = to_tuf_datetime_format(start_date, interval)
    targets_metadata_path = Path(taf_happy_path.metadata_path) / "targets.json"
    snapshot_metadata_path = Path(taf_happy_path.metadata_path) / "snapshot.json"
    old_targets_metadata = targets_metadata_path.read_bytes()
    old_snapshot_metadata = snapshot_metadata_path.read_bytes()
    taf_happy_path.update_timestamp_keystores(
        [timestamp_key], start_date=start_date, interval=interval
    )
    new_timestamp_metadata = str(Path(taf_happy_path.metadata_path) / "timestamp.json")
    signable = securesystemslib.util.load_json_file(new_timestamp_metadata)
    tuf.formats.SIGNABLE_SCHEMA.check_match(signable)
    actual_expiration_date = signable["signed"]["expires"]

    assert actual_expiration_date == expected_expiration_date
    # check if targets and snapshot remained the same
    assert old_targets_metadata == targets_metadata_path.read_bytes()
    assert old_snapshot_metadata == snapshot_metadata_path.read_bytes()


def test_update_timestamp_wrong_key(repositories, snapshot_key):
    taf_happy_path = repositories["test-happy-path"]
    with pytest.raises(taf.exceptions.InvalidKeyError):
        taf_happy_path.update_timestamp_keystores([snapshot_key])


def test_update_targets_from_keystore_valid_key(repositories, targets_key):
    taf_happy_path = repositories["test-happy-path"]
    targets_path = Path(taf_happy_path.targets_path)
    repositories_json_path = targets_path / "repositories.json"

    branch_id = "14e81cd1-0050-43aa-9e2c-e34fffa6f517"
    target_commit_sha = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
    repositories_json_old = repositories_json_path.read_text()

    targets_data = {
        "branch": {"target": branch_id},
        "dummy/target_dummy_repo": {"target": {"commit": target_commit_sha}},
        "test_file": {},
    }

    taf_happy_path.update_targets_keystores(
        [targets_key],
        added_targets_data=targets_data,
        start_date=datetime.datetime.now(),
    )

    assert (targets_path / "branch").read_text() == branch_id
    assert target_commit_sha in (targets_path / "dummy/target_dummy_repo").read_text()
    assert (targets_path / "test_file").is_file()
    assert repositories_json_old == repositories_json_path.read_text()


def test_update_targets_from_keystore_wrong_key(repositories, snapshot_key):
    taf_happy_path = repositories["test-happy-path"]
    targets_data = {"test_file": {}}

    with pytest.raises(taf.exceptions.TargetsMetadataUpdateError):
        taf_happy_path.update_targets_keystores([snapshot_key], targets_data)


def test_update_targets_valid_key_valid_pin(repositories, targets_yk):
    taf_happy_path = repositories["test-happy-path"]
    if targets_yk.scheme != DEFAULT_RSA_SIGNATURE_SCHEME:
        pytest.skip()
    targets_path = Path(taf_happy_path.targets_path)
    repositories_json_path = targets_path / "repositories.json"

    branch_id = "14e81cd1-0050-43aa-9e2c-e34fffa6f517"
    target_commit_sha = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
    repositories_json_old = repositories_json_path.read_text()

    targets_data = {
        "branch": {"target": branch_id},
        "dummy/target_dummy_repo": {"target": {"commit": target_commit_sha}},
        "test_file": {},
    }
    yk.add_key_pin(targets_yk.serial, VALID_PIN)
    targets_yk.insert()
    public_key = targets_yk.tuf_key
    taf_happy_path.update_targets_yubikeys(
        [public_key],
        added_targets_data=targets_data,
        start_date=datetime.datetime.now(),
    )

    assert (targets_path / "branch").read_text() == branch_id
    assert target_commit_sha in (targets_path / "dummy/target_dummy_repo").read_text()
    assert (targets_path / "test_file").is_file()
    assert repositories_json_old == repositories_json_path.read_text()


def test_delete_target_file_valid_key_valid_pin(repositories, targets_yk):
    taf_happy_path = repositories["test-happy-path"]
    if targets_yk.scheme != DEFAULT_RSA_SIGNATURE_SCHEME:
        pytest.skip()
    targets_path = Path(taf_happy_path.targets_path)

    yk.add_key_pin(targets_yk.serial, VALID_PIN)
    targets_yk.insert()
    public_key = targets_yk.tuf_key

    # add test_file
    targets_data = {"test_file": {}}
    taf_happy_path.update_targets_yubikeys(
        [public_key],
        added_targets_data=targets_data,
        start_date=datetime.datetime.now(),
    )

    assert (targets_path / "test_file").is_file()
    targets_obj = taf_happy_path._role_obj("targets")
    assert "test_file" in targets_obj.target_files

    # remove test_file
    taf_happy_path.update_targets_yubikeys(
        [public_key],
        removed_targets_data=targets_data,
        start_date=datetime.datetime.now(),
    )

    assert not (targets_path / "test_file").is_file()
    targets_obj = taf_happy_path._role_obj("targets")
    assert "test_file" not in targets_obj.target_files


@pytest.mark.skipif(TEST_WITH_REAL_YK, reason="Testing with real Yubikey.")
def test_update_targets_wrong_key(repositories, root1_yk):
    taf_happy_path = repositories["test-happy-path"]
    targets_data = {"test_file": {}}

    with pytest.raises(taf.exceptions.TargetsMetadataUpdateError):
        root1_yk.insert()
        yk.add_key_pin(root1_yk.serial, VALID_PIN)
        taf_happy_path.update_targets_yubikeys(
            [root1_yk.tuf_key], added_targets_data=targets_data
        )



def test_get_all_targets_roles(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]
    assert taf_delegated_roles.get_all_targets_roles() == [
        "targets",
        "delegated_role1",
        "delegated_role2",
        "inner_delegated_role",
    ]


def test_find_roles_parent(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]
    assert (
        taf_delegated_roles.find_delegated_roles_parent("delegated_role1") == "targets"
    )
    assert (
        taf_delegated_roles.find_delegated_roles_parent("delegated_role2") == "targets"
    )
    assert (
        taf_delegated_roles.find_delegated_roles_parent("inner_delegated_role")
        == "delegated_role2"
    )


def test_map_signing_roles(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]
    expected_targets_roles = {
        "dir1/delegated_role1_1.txt": "delegated_role1",
        "dir1/delegated_role1_2.txt": "delegated_role1",
        "dir2/delegated_role2_1.txt": "delegated_role2",
        "dir2/delegated_role2_1.txt": "delegated_role2",
        "dir2/inner_delegated_role.txt": "inner_delegated_role",
    }
    actual_targets_roles = taf_delegated_roles.map_signing_roles(
        expected_targets_roles.keys()
    )
    for file_name, expected_role in expected_targets_roles.items():
        assert file_name in actual_targets_roles
        assert actual_targets_roles[file_name] == expected_role


def test_find_keys_roles(
    repositories,
    delegated_role11_key,
    delegated_role12_key,
    delegated_role13_key,
    delegated_role2_key,
    inner_delegated_role_key,
):
    taf_delegated_roles = repositories["test-delegated-roles"]
    assert not len(taf_delegated_roles.find_keys_roles([delegated_role11_key]))
    assert taf_delegated_roles.find_keys_roles(
        [delegated_role11_key, delegated_role12_key]
    ) == ["delegated_role1"]
    assert taf_delegated_roles.find_keys_roles(
        [delegated_role11_key, delegated_role12_key, delegated_role13_key]
    ) == ["delegated_role1"]
    assert taf_delegated_roles.find_keys_roles([delegated_role2_key]) == [
        "delegated_role2"
    ]
    assert taf_delegated_roles.find_keys_roles([inner_delegated_role_key]) == [
        "inner_delegated_role"
    ]


def test_sort_roles_targets_for_filenames(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]
    targets_files_by_roles = taf_delegated_roles.sort_roles_targets_for_filenames()
    assert "delegated_role1" in targets_files_by_roles
    for target_file in targets_files_by_roles["delegated_role1"]:
        assert target_file in [
            "dir1/delegated_role1_1.txt",
            "dir1/delegated_role1_2.txt",
        ]

    assert "delegated_role2" in targets_files_by_roles
    for target_file in targets_files_by_roles["delegated_role2"]:
        assert target_file in [
            "dir2/delegated_role2_1.txt",
            "dir2/delegated_role2_2.txt",
        ]
    assert "inner_delegated_role" in targets_files_by_roles
    assert targets_files_by_roles["inner_delegated_role"] == [
        "dir2/inner_delegated_role.txt"
    ]


def test_get_role_from_target_paths(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]

    target_paths_with_same_role = [
        "dir1/delegated_role1_1.txt",
        "dir1/delegated_role1_2.txt",
    ]
    role = taf_delegated_roles.get_role_from_target_paths(target_paths_with_same_role)
    assert role == "delegated_role1"

    target_paths_with_different_role = [
        "dir2/delegated_role2_1.txt",
        "dir2/inner_delegated_role.txt",
    ]
    role = taf_delegated_roles.get_role_from_target_paths(
        target_paths_with_different_role
    )
    assert role is None



import pytest
import taf.repositoriesdb as repositoriesdb
from taf.auth_repo import AuthenticationRepo
import taf.settings as settings
from taf.tests.conftest import load_repositories

AUTH_REPO_NAME = "organization/auth_repo"


def setup_module(module):
    settings.update_from_filesystem = True


def teardown_module(module):
    settings.update_from_filesystem = False


@pytest.mark.parametrize(
    "test_name",
    [
        "test-no-delegations",
        "test-delegated-roles",
        "test-delegated-roles-with-mirrors",
    ],
)
def test_load_repositories(test_name, repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories[test_name]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    with load_repositories(auth_repo):
        _check_repositories_dict(repositories, auth_repo, auth_repo.head_commit_sha())


@pytest.mark.parametrize("test_name", ["test-no-delegations", "test-delegated-roles"])
def test_load_repositories_only_load_targets(
    test_name, repositoriesdb_test_repositories
):
    repositories = repositoriesdb_test_repositories[test_name]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    with load_repositories(auth_repo, only_load_targets=True):
        _check_repositories_dict(
            repositories, auth_repo, auth_repo.head_commit_sha(), only_load_targets=True
        )


def test_load_repositories_of_roles(repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories["test-delegated-roles"]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    roles = ["delegated_role1"]
    with load_repositories(auth_repo, roles=roles):
        _check_repositories_dict(
            repositories, auth_repo, auth_repo.head_commit_sha(), roles=roles
        )


@pytest.mark.parametrize("test_name", ["test-no-delegations", "test-delegated-roles"])
def test_load_repositories_all_commits(test_name, repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories[test_name]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    commits = auth_repo.all_commits_on_branch()[1:]  # remove the first commit
    with load_repositories(auth_repo, commits=commits):
        _check_repositories_dict(repositories, auth_repo, *commits)


def test_get_deduplicated_repositories(repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories["test-delegated-roles"]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    commits = auth_repo.all_commits_on_branch()[1:]  # remove the first commit
    with load_repositories(auth_repo, commits=commits):
        repos = repositoriesdb.get_deduplicated_repositories(auth_repo, commits)
        assert len(repos) == 3
        for repo_name in repositories:
            if repo_name != AUTH_REPO_NAME:
                assert repo_name in repos


def test_get_repository(repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories["test-delegated-roles"]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    commits = auth_repo.all_commits_on_branch()[1:]  # remove the first commit
    path = "namespace/TargetRepo1"
    with load_repositories(auth_repo, commits=commits):
        repo = repositoriesdb.get_repository(auth_repo, path, commits[-1])
        assert repo.name == path
        repo = repositoriesdb.get_repository(auth_repo, path, commits[-2])
        assert repo.name == path


def test_get_repository_by_custom_data(repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories["test-delegated-roles"]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    with load_repositories(auth_repo):
        for repo_type, repo_name in [
            ("type1", "namespace/TargetRepo1"),
            ("type2", "namespace/TargetRepo2"),
            ("type3", "namespace/TargetRepo3"),
        ]:
            type_repos = repositoriesdb.get_repositories_by_custom_data(
                auth_repo, type=repo_type
            )
            assert len(type_repos) == 1
            assert type_repos[0].name == repo_name


def test_get_repositories_paths_by_custom_data(repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories["test-delegated-roles"]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    with load_repositories(auth_repo):
        for repo_type, repo_name in [
            ("type1", "namespace/TargetRepo1"),
            ("type2", "namespace/TargetRepo2"),
            ("type3", "namespace/TargetRepo3"),
        ]:
            paths = repositoriesdb.get_repositories_paths_by_custom_data(
                auth_repo, type=repo_type
            )
            assert paths == [repo_name]


def _check_repositories_dict(
    repositories, auth_repo, *commits, roles=None, only_load_targets=False
):
    assert auth_repo.name in repositoriesdb._repositories_dict
    auth_repos_dict = repositoriesdb._repositories_dict[auth_repo.name]
    if roles is not None and len(roles):
        only_load_targets = True
    if only_load_targets:
        target_files_of_roles = auth_repo.get_singed_target_files_of_roles(roles)
    for commit in commits:
        repositories_json = auth_repo.get_json(
            commit, repositoriesdb.REPOSITORIES_JSON_PATH
        )
        repositories_data = repositories_json["repositories"]
        assert commit in auth_repos_dict
        for repo_name in repositories:
            if repo_name != AUTH_REPO_NAME:
                if not only_load_targets or (
                    only_load_targets and repo_name in target_files_of_roles
                ):
                    assert repo_name in auth_repos_dict[commit]
                    # check custom data
                    custom_data = repositories_data[repo_name].get("custom", {})
                    assert (
                        auth_repos_dict[commit][repo_name].additional_info
                        == custom_data
                    )
                else:
                    assert repo_name not in auth_repos_dict[commit]



import json
import os
import tempfile
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from tuf.repository_tool import METADATA_DIRECTORY_NAME
from taf.git import GitRepository, NamedGitRepository
from taf.repository_tool import (
    Repository as TAFRepository,
    get_role_metadata_path,
    get_target_path,
)


class AuthRepoMixin(TAFRepository):

    LAST_VALIDATED_FILENAME = "last_validated_commit"
    TEST_REPO_FLAG_FILE = "test-auth-repo"

    @property
    def conf_dir(self):
        """
        Returns location of the directory which stores the authentication repository's
        configuration files. That is, the last validated commit.
        Create the directory if it does not exist.
        """
        # the repository's name consists of the namespace and name (namespace/name)
        # the configuration directory should be _name
        last_dir = os.path.basename(os.path.normpath(self.path))
        conf_path = Path(self.path).parent / f"_{last_dir}"
        conf_path.mkdir(parents=True, exist_ok=True)
        return str(conf_path)

    @property
    def certs_dir(self):
        certs_dir = Path(self.path, "certs")
        certs_dir.mkdir(parents=True, exist_ok=True)
        return str(certs_dir)

    @property
    def last_validated_commit(self):
        """
        Return the last validated commit of the authentication repository
        """
        try:
            return Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).read_text()
        except FileNotFoundError:
            return None

    @property
    def log_prefix(self):
        return f"Auth repo {self.name}: "

    def get_target(self, target_name, commit=None, safely=True):
        if commit is None:
            commit = self.head_commit_sha()
        target_path = get_target_path(target_name)
        if safely:
            return self.safely_get_json(commit, target_path)
        else:
            return self.get_json(commit, target_path)

    def is_commit_authenticated(self, target_name, commit):
        """Checks if passed commit is ever authenticated for given target name.
        """
        for auth_commit in self.all_commits_on_branch(reverse=False):
            target = self.get_target(target_name, auth_commit)
            try:
                if target["commit"] == commit:
                    return True
            except TypeError:
                continue
        return False

    @contextmanager
    def repository_at_revision(self, commit):
        """
        Context manager which makes sure that TUF repository is instantiated
        using metadata files at the specified revision. Creates a temp directory
        and metadata files inside it. Deleted the temp directory when no longer
        needed.
        """
        tuf_repository = self._tuf_repository
        with tempfile.TemporaryDirectory() as temp_dir:
            metadata_files = self.list_files_at_revision(
                commit, METADATA_DIRECTORY_NAME
            )
            Path(temp_dir, METADATA_DIRECTORY_NAME).mkdir(parents=True)
            for file_name in metadata_files:
                path = Path(temp_dir, METADATA_DIRECTORY_NAME, file_name)
                with open(path, "w") as f:
                    data = self.get_json(
                        commit, f"{METADATA_DIRECTORY_NAME}/{file_name}"
                    )
                    json.dump(data, f)
            self._load_tuf_repository(temp_dir)
            yield
            self._tuf_repository = tuf_repository

    def set_last_validated_commit(self, commit):
        """
        Set the last validated commit of the authentication repository
        """
        self._log_debug(f"setting last validated commit to: {commit}")
        Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).write_text(commit)

    def sorted_commits_and_branches_per_repositories(self, commits, target_repos=None):
        """Return a dictionary consisting of branches and commits belonging
        to it for every target repository:
        {
            target_repo1: {
                branch1: [{"commit": commit1, "additional-info": {...}}, {"commit": commit2, "additional-info": {...}}, {"commit": commit3, "additional-info": {}}],
                branch2: [{"commit": commit4, "additional-info: {...}}, {"commit": commit5, "additional_info": {}}]
            },
            target_repo2: {
                branch1: [{"commit": commit6, "additional-info": {...}}],
                branch2: [{"commit": commit7", "additional-info": {...}]
            }
        }
        Keep in mind that targets metadata
        file is not updated everytime something is committed to the authentication repo.
        """
        repositories_commits = defaultdict(dict)
        targets = self.targets_at_revisions(*commits, target_repos=target_repos)
        previous_commits = {}
        for commit in commits:
            for target_path, target_data in targets[commit].items():
                target_branch = target_data.get("branch")
                target_commit = target_data.get("commit")
                previous_commit = previous_commits.get(target_path)
                if previous_commit is None or target_commit != previous_commit:
                    repositories_commits[target_path].setdefault(
                        target_branch, []
                    ).append(
                        {
                            "commit": target_commit,
                            "additional-info": target_data.get("additional-info"),
                        }
                    )
                previous_commits[target_path] = target_commit
        self._log_debug(
            f"new commits per repositories according to target files: {repositories_commits}"
        )
        return repositories_commits

    def targets_at_revisions(self, *commits, target_repos=None):
        targets = defaultdict(dict)
        for commit in commits:
            # repositories.json might not exit, if the current commit is
            # the initial commit
            repositories_at_revision = self.safely_get_json(
                commit, get_target_path("repositories.json")
            )
            if repositories_at_revision is None:
                continue
            repositories_at_revision = repositories_at_revision["repositories"]

            # get names of all targets roles defined in the current revision
            with self.repository_at_revision(commit):
                roles_at_revision = self.get_all_targets_roles()

            for role_name in roles_at_revision:
                # targets metadata files corresponding to the found roles must exist
                targets_at_revision = self.get_json(
                    commit, get_role_metadata_path(role_name)
                )
                targets_at_revision = targets_at_revision["signed"]["targets"]

                for target_path in targets_at_revision:
                    if target_path not in repositories_at_revision:
                        # we only care about repositories
                        continue
                    if target_repos is not None and target_path not in target_repos:
                        # if specific target repositories are specified, skip all other
                        # repositories
                        continue
                    target_content = self.safely_get_json(
                        commit, get_target_path(target_path)
                    )
                    if target_content is not None:
                        target_commit = target_content.pop("commit")
                        target_branch = target_content.pop("branch", "master")
                        targets[commit][target_path] = {
                            "branch": target_branch,
                            "commit": target_commit,
                            "additional-info": target_content,
                        }
        return targets


class AuthenticationRepo(GitRepository, AuthRepoMixin):
    def __init__(
        self,
        path,
        repo_urls=None,
        additional_info=None,
        default_branch="master",
        *args,
        **kwargs,
    ):
        super().__init__(path, repo_urls, additional_info, default_branch)


class NamedAuthenticationRepo(NamedGitRepository, AuthRepoMixin):
    def __init__(
        self,
        root_dir,
        repo_name,
        repo_urls=None,
        additional_info=None,
        default_branch="master",
        *args,
        **kwargs,
    ):
        super().__init__(
            root_dir=root_dir,
            repo_name=repo_name,
            repo_urls=repo_urls,
            additional_info=additional_info,
            default_branch=default_branch,
        )



import taf.repositoriesdb as repositoriesdb
from taf.auth_repo import AuthenticationRepo
from taf.tests.conftest import load_repositories

AUTH_REPO_NAME = "organization/auth_repo"

REPOS_URLS = {
    f"namespace/{repo_name}": [
        f"https://github.com/namespace/{repo_name}.git",
        f"https://github.com/test/namespace-{repo_name}.git",
        f"https://gitlab.com/namespace2/{repo_name}.git",
        f"https://gitlab.com/namespace/namespace--{repo_name}.git",
        f"git@github.com:namespace/{repo_name}.git",
    ]
    for repo_name in ["TargetRepo1", "TargetRepo2", "TargetRepo3"]
}


def test_load_repositories_with_mirrors(repositoriesdb_test_repositories):
    repositories = repositoriesdb_test_repositories["test-delegated-roles-with-mirrors"]
    auth_repo = AuthenticationRepo(repositories[AUTH_REPO_NAME])
    commit = auth_repo.head_commit_sha()
    with load_repositories(auth_repo):
        for repo_path in repositories:
            loaded_repos_dict = repositoriesdb._repositories_dict[auth_repo.name][
                commit
            ]
            if repo_path != AUTH_REPO_NAME:
                repo = loaded_repos_dict[repo_path]
                assert repo.repo_urls == REPOS_URLS[repo_path]



import json
import os
from pathlib import Path

import pytest
from pytest import fixture

from taf.exceptions import TargetsError
from taf.git import GitRepository


@fixture(autouse=True)
def run_around_tests(repositories):
    yield
    for taf_repository in repositories.values():
        repo = GitRepository(taf_repository.path)
        repo.reset_to_head()
        repo.clean()
        taf_repository._repository.targets.clear_targets()


def test_add_targets_new_files(repositories):
    taf_happy_path = repositories["test-happy-path"]
    old_targets = {"targets": _get_old_targets(taf_happy_path)}

    json_file_content = {"attr1": "value1", "attr2": "value2"}
    regular_file_content = "this file is not empty"
    data = {
        "new_json_file": {"target": json_file_content},
        "new_file": {"target": regular_file_content},
        "empty_file": {"target": None},
    }
    role = taf_happy_path.modify_targets(data)
    assert role == "targets"

    _check_target_files(taf_happy_path, data, old_targets)


def test_add_targets_nested_files(repositories):
    taf_happy_path = repositories["test-happy-path"]
    old_targets = {"targets": _get_old_targets(taf_happy_path)}

    data = {
        "inner_folder1/new_file_1": {"target": "file 1 content"},
        "inner_folder2/new_file_2": {"target": "file 2 content"},
    }
    taf_happy_path.modify_targets(data)
    _check_target_files(taf_happy_path, data, old_targets)


def test_add_targets_files_to_keep(repositories):
    taf_happy_path = repositories["test-happy-path"]
    old_targets = {"targets": _get_old_targets(taf_happy_path)}
    data = {"a_new_file": {"target": "new file content"}}
    taf_happy_path.modify_targets(data)
    _check_target_files(taf_happy_path, data, old_targets, files_to_keep=["branch"])


def test_add_targets_delegated_roles_no_child_roles(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]
    old_targets = {
        "delegated_role1": ["dir1/delegated_role1_1.txt", "dir1/delegated_role1_2.txt"],
        "delegated_role2": ["dir2/delegated_role2_1.txt", "dir2/delegated_role2_2.txt"],
        "inner_delegated_role": ["dir2/inner_delegated_role.txt"],
    }

    data = {"dir1/a_new_file": {"target": "new file content"}}
    role = taf_delegated_roles.modify_targets(data)
    assert role == "delegated_role1"

    _check_target_files(taf_delegated_roles, data, old_targets, role)


def test_add_targets_multiple_delegated_roles_should_raise_error(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]
    data = {
        "dir1/a_new_file": {"target": "new file content"},
        "dir2/a_new_file": {"target": "new file content"},
    }

    with pytest.raises(TargetsError):
        taf_delegated_roles.modify_targets(data)


def test_add_targets_delegated_roles_child_roles(repositories):
    taf_delegated_roles = repositories["test-delegated-roles"]
    old_targets = {
        "delegated_role1": ["dir1/delegated_role1_1.txt", "dir1/delegated_role1_2.txt"],
        "delegated_role2": ["dir2/delegated_role2_1.txt", "dir2/delegated_role2_2.txt"],
        "inner_delegated_role": ["dir2/inner_delegated_role.txt"],
    }

    data = {"dir2/a_new_file": {"target": "new file content"}}
    role = "delegated_role2"
    taf_delegated_roles.modify_targets(data)
    _check_target_files(taf_delegated_roles, data, old_targets, role)


def _check_target_files(
    repo, data, old_targets, targets_role="targets", files_to_keep=None
):
    if files_to_keep is None:
        files_to_keep = []

    targets_path = repo.targets_path
    for target_rel_path, content in data.items():
        target_path = targets_path / target_rel_path
        assert target_path.exists()
        with open(str(target_path)) as f:
            file_content = f.read()
            target_content = content["target"]
            if isinstance(target_content, dict):
                content_json = json.loads(file_content)
                assert content_json == target_content
            elif target_content:
                assert file_content == target_content
            else:
                assert file_content == ""

    # make sure that everything defined in repositories.json still exists if there was
    # repositories.json
    repository_targets = []
    for _, roles_targets in old_targets.items():
        if "repositories.json" in roles_targets:
            repositories_path = targets_path / "repositories.json"
            assert repositories_path.exists()
            with open(str(repositories_path)) as f:
                repositories = json.load(f)["repositories"]
                for target_rel_path in repositories:
                    target_path = targets_path / target_rel_path
                    assert target_path.exists()
                    repository_targets.append(target_rel_path)
            break

    # make sure that files to keep exist
    for file_to_keep in files_to_keep:
        # if the file didn't exists prior to adding new targets
        # it won't exists after adding them
        if file_to_keep not in old_targets:
            continue
        target_path = targets_path / file_to_keep
        assert target_path.exists()


def _get_old_targets(repo):
    targets_path = repo.targets_path
    old_targets = []
    for root, _, filenames in os.walk(str(targets_path)):
        for filename in filenames:
            rel_path = Path(root, filename).relative_to(targets_path)
            old_targets.append(rel_path.as_posix())
    return old_targets
